/**
 * Name: thermocouple_example.c
 * Desc: Demonstrates thermocouple configuration and  measurement
 *       https://labjack.com/support/app-notes/thermocouples
 *
 * Relevant Documentation:
 *
 * LJM Library:
 *	LJM Library Installer:
 *		https://labjack.com/support/software/installers/ljm
 *	LJM Users Guide:
 *		https://labjack.com/support/software/api/ljm
 *	Opening and Closing:
 *		https://labjack.com/support/software/api/ljm/function-reference/opening-and-closing
 *	Single Value Functions(such as eReadName):
 *		https://labjack.com/support/software/api/ljm/function-reference/single-value-functions
 *	TCVoltsToTemp:
 *		https://labjack.com/support/software/api/ud/function-reference/tcvoltstotemp
 *
 * T-Series and I/O:
 *	Modbus Map:
 *		https://labjack.com/support/software/api/modbus/modbus-map
 *	Analog Inputs:
 *		https://labjack.com/support/datasheets/t-series/ain
**/

// For printf
#include <stdio.h>

// For the LabJackM Library
#include <LabJackM.h>

// For LabJackM helper functions, such as OpenOrDie, PrintDeviceInfoFromHandle,
// ErrorCheck, etc.
#include "../../LJM_Utilities.h"

enum {DEGK, DEGC, DEGF};

struct TCData {
	long tcType;
	int posChannel;
	int negChannel;
	unsigned short tempUnits;
	int gain;
	float offset;
};

void setupThermocouple(int handle, int deviceType, struct TCData tcData) {
	int err;
	// For converting TC type constant to T7 TC AIN_EF index (T7 only)
	const int TC_INDEX_LUT[9] = {28,20,21,22,27,23,25,24,30};
	const int TC_INDEX_OFFSET = 6000;

	// Set the resolution index to the highest available setting
	err = LJM_eWriteAddress(handle, 41500+tcData.posChannel, LJM_UINT16, 0);
	ErrorCheck(err, "setupThermocouple:Setting AIN resolution index");

	if(deviceType == LJM_dtT7) {
		// Set the negative channel
		err = LJM_eWriteAddress(handle,
								41000+tcData.posChannel,
								LJM_UINT16,
								tcData.negChannel);
		ErrorCheck(err, "setupThermocouple:Setting AIN negative channel");
		// Set up the T7 AIN#_EF_INDEX
		err = LJM_eWriteAddress(handle,
								9000+2*tcData.posChannel,
								LJM_UINT32,
								TC_INDEX_LUT[tcData.tcType - TC_INDEX_OFFSET]);
		ErrorCheck(err, "setupThermocouple:Setting AIN_EF Index");
		// Set up the T7 AIN#_EF_CONFIG_A (temperature units to degK)
		err = LJM_eWriteAddress(handle,
								9300+2*tcData.posChannel, 
								LJM_UINT32, 
								DEGK);
		ErrorCheck(err, "setupThermocouple:Setting AIN_EF ConfigA");
		
	}
}

void printThermocoupleMeas(int handle, int deviceType, struct TCData tcData) {
	double TCVolts, CJTempK, TCTempK, TempC, TempF;
	// We expect the cold junction to be ~3 degrees cooler than the internal temp sensor.
	const double CJ_TEMP_CORRECTION = -3.0;
	int err;

	if(deviceType == LJM_dtT4) {
		err = LJM_eReadAddress(handle, tcData.posChannel, LJM_FLOAT32, &TCVolts);
		ErrorCheck(err, "Reading TC AIN");

		// Account for Tick-InAmp scaling
		TCVolts = (TCVolts - tcData.offset) / tcData.gain;

		err = LJM_eReadAddress(handle, 60052, LJM_FLOAT32, &CJTempK);
		ErrorCheck(err, "Reading internal temp sensor");

		err = LJM_TCVoltsToTemp(LJM_ttK, TCVolts, CJTempK + CJ_TEMP_CORRECTION, &TCTempK);
		ErrorCheck(err, "Calculating TCTempK");
	}

	else if(deviceType == LJM_dtT7) {
		err = LJM_eReadAddress(handle, 7300+2*tcData.posChannel, LJM_FLOAT32, &TCVolts);
		ErrorCheck(err, "Reading TC Volts");

		err = LJM_eReadAddress(handle, 7600+2*tcData.posChannel, LJM_FLOAT32, &CJTempK);
		ErrorCheck(err, "Reading CJC temperature");

		err = LJM_eReadAddress(handle, 7000+2*tcData.posChannel, LJM_FLOAT32, &TCTempK);
		ErrorCheck(err, "Reading TC Temperature");
	}

	printf("TCVolts: %lf,\tCJTempK: %lf,\t", TCVolts, CJTempK);
	TempC = TCTempK-273.15;
	TempF = (1.8*TCTempK)-459.67;

	switch(tcData.tempUnits){
		case DEGK:
			printf("deg K: %lf\n", TCTempK);
		break;

		case DEGC:
			printf("deg C: %lf\n", TempC);
		break;

		case DEGF:
			
			printf("deg F: %lf\n", TempF);
		break;
		default:
			printf("deg K: %lf\n", TCTempK);
		break;
	}
}

int main()
{
	int err, handle, i, deviceType, connectionType, serialNumber;
	int ipAddress, portOrPipe, packetMaxBytes;
	struct TCData tcData;

	// Supported TC types are:
	//     LJM_ttB
	//     LJM_ttE
	//     LJM_ttJ
	//     LJM_ttK
	//     LJM_ttN
	//     LJM_ttR
	//     LJM_ttS
	//     LJM_ttT
	//     LJM_ttC
	tcData.tcType = LJM_ttK;

	// If taking a differential reading on a T7, this should be the even
	// numbered AIN channel, and the negative channel should be connected to
	// the positive AIN channel plus one. ex: posChannel=AIN0, negChannel=AIN1
	tcData.posChannel = 0; //AIN0

	// Value of 199 indicates single ended measurement
	// Config ignored for the T4
	tcData.negChannel = 199;

	// Set the temperature units to use
	tcData.tempUnits = DEGC;

	// Gain set for Tick-InAmp (for T4). Config ignored for T7
	tcData.gain = 51;

	// Offset voltage setting on Tick-InAmp (for T4). Config ignored for T7
	tcData.offset = 0.4;

	// Open first found LabJack
	handle = OpenOrDie(LJM_dtANY, LJM_ctANY, "LJM_idANY");
	// handle = OpenSOrDie("LJM_dtANY", "LJM_ctANY", "LJM_idANY");

	// Get device info
	err = LJM_GetHandleInfo(handle, &deviceType, &connectionType, &serialNumber, &ipAddress,
		&portOrPipe, &packetMaxBytes);
	ErrorCheck(err, "LJM_GetHandleInfo");

	PrintDeviceInfo(deviceType, connectionType, serialNumber, ipAddress, portOrPipe, packetMaxBytes);

	setupThermocouple(handle, deviceType, tcData);


	printf("\nPress ctrl + c to stop\n");
	for (i=0; ; i++) {
		printThermocoupleMeas(handle, deviceType, tcData);
		MillisecondSleep(1000);
	}

	WaitForUserIfWindows();

	return LJME_NOERROR;
}
